/* authlocal.c - auto-authenticate users connecting from localhost.
 *
 *	The procedure you want to call from your code is getUID,
 *	which returns a uid_t, or 0 if authentication failed.
 *	Be sure to #define USE_AUTHLOCAL if you want it to work!
 *
 *	Remember that uid's under 100 won't be recognized; this is because
 *	they're normally system daemon accounts, not real users.  If you
 *	need to change this behavior, change AUTHLOCAL_MIN_UID in authlocal.h
 *	before compiling.
 *
 * jer 19990907.1515	Jeremy Monin <jeremy@shadowlands.org>
 *	http://shadowlands.org/authlocal/
 */

#include <unistd.h>  // for ssize_t, uid_t
#include <stdio.h>
#include <stdlib.h>  // for atol
#include <sys/types.h>  // for open
#include <sys/stat.h>   // for open
#include <fcntl.h>      // for open
#include <string.h>  // for memmove
#include <sys/socket.h>  // this is
#include <netinet/in.h>  // all for
#include <arpa/inet.h>   // inet_addr

#include "authlocal.h"

/*
 * How it Works:
 *
 * When connecting to a TCP server from localhost, the server knows
 * the arbitrary port# you're connecting from... under linux, the file
 * /proc/net/tcp contains data about this socket, including its owner.
 * This can be used to automatically, securely authenticate.
 * See an example at the bottom of this file.
 *
 * It does lead to the problem of not _wanting_ to log in as yourself,
 * but this is a separate issue.  Also, be sure root doesn't log in this way.
 * We also should check to be sure the logging-in user has a valid shell,
 * but we don't right now.
 *
 * Tested and successful under linux 2.0.x and 2.2.x
 *
 * Be sure to define USE_AUTHLOCAL if you want to use this;
 * if this isn't defined, the entire getUID function will ALWAYS
 * return 0 (failure).  This means you don't need to change any of
 * your other code, assuming you have a fallback method for authenticating
 * people, which you'll need anyway if you want them to be able to
 * get in from other hosts.
 */

/*
    authlocal.c version 1.0 - Automatic authentication of connections
    from localhost.
    Copyright (C) 2000 Jeremy D. Monin (jeremy@shadowlands.org).

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/



#ifdef USE_AUTHLOCAL

	/* yes, most of the file's in this ifdef. */
	/* Only thing that isn't is, if USE_AUTHLOCAL
	 * isn't defined, the function that gets called
	 * from outside this module always returns 0 (failure),
	 * meaning you should fall back to standard username-and-pw-prompt
	 * authentication.
	 */


/* format of /proc/net/tcp lines: (as of linux v2.x, not sure of earlier vers)
 * We want to get the UID.
 *    sl    addy       addy   st dunno:d   d:dunno uh uid timeout inode
 * \n  n: local:port rem:port nn nnnn:nnn nn:nnnn nnn uid  n n\n
 *        ^^^^^^^^    ^^^^^  in hex; 8, 4 chars respectively
 */

#define PATSIZE    (8+1+4+1+8+1+4+1)
#define LINEBUFSIZ 4096
#define TCPSTAT_FNAME "/proc/net/tcp"


/* check to see if we need to read more chars into the buffer;
 * if so, do that and adjust the end-of-buffer-data pointer.
 * Either way, return it.
 */
static char *read_chunk (char *buf, char **c, char *bufend, int inputfd) {
  ssize_t nread;
  int inbufloc = *c - buf;  /* how far into buffer are we? */


  /* Must read some into the buffer...  First, move current buffer
   * contents up so we have room for more.
   */
  memmove (buf, *c, LINEBUFSIZ - inbufloc);
  *c = buf + (LINEBUFSIZ - inbufloc);
  **c = '\0';    /* *c now points to where new stuff should be added */

  /* so add the new stuff... */
  nread = read (inputfd, buf, LINEBUFSIZ - (*c - buf) - 1);
  if (nread == -1) {
    perror ("read");
    return NULL;
  }
  if (nread == 0)    /* EOF */
    return NULL;

  /* adjust c to start of buf, adjust end of buf ptr */
  nread += (*c - buf);
  *(buf + nread) = '\0';
  *c = buf;

  return (buf + nread);
}

static char *buf_begin_get(char *buf, int readfromfd) {
  ssize_t nread = read (readfromfd, buf, LINEBUFSIZ-1);
  if (nread == -1) {
    return NULL;
  }

  *(buf + nread) = '\0';
  return (buf + nread);
}



/* returns 0 if something goes wrong and it can't be found.
 * (UID 0 really shouldn't be logging in from a network anyway.)
 */
static uid_t getUID_impl (unsigned int /* type? */ localhost_ip,
              int /* type? */ userportno, int daemonportno,
              int inputfd) {

  char wantthis_in_hex[PATSIZE];
  char buf[LINEBUFSIZ];
  char *b;
  char *bufend;
  char *c;

  sprintf (wantthis_in_hex, "%08X:%04X %08X:%04X",
	   localhost_ip, (int) userportno, localhost_ip, (int) daemonportno);

  /* search for a line containing this */
  bufend = buf_begin_get (buf, inputfd);
  if (bufend == NULL) {
    return (uid_t) 0;
  }


  for ( c = buf;
	(bufend != NULL) && (*c != '\0') && (c < bufend);
	bufend = read_chunk (buf, &c, bufend, inputfd)) {
 
    /* get c to point to the string we seek  */
    b = c;
    c = strstr (c, wantthis_in_hex);

    /* skip to next if not the right one (not on this line) */
    if (c == NULL) {  /* not found, get more to search through */
      c = rindex (b, '\n');
      if (c == NULL)
	c = bufend;
    } else {
      /* now get uid and return it; it's 3rd field before end of line */
      int bkfld;

      /* point c at end-of-line */
      c = index (c, '\n');
      if (c == NULL)
	c = bufend;

      for (bkfld = 1; bkfld <= 3; ++bkfld) {
	/* now inch c backward through whitespace */
	while (*c <= ' ')
	  --c;
	if (bkfld == 3) /* we're right at end of UID; null-terminate it */
	  *(c+1) = '\0';
	
	/* this is the field data; inch backward to next whitespace */
	while (*c > ' ')
	  --c;
      }  /* end moving through loop */
	    
      /* now nudge forward and read it!  have a nice day,
       * we're done.
       */
      ++c;
      /* read it */
      return (uid_t) atol(c);
    }  // end if
  }  // end loop
  return (uid_t) 0;  /* couldn't find it. */
}




#endif	/* the big IFDEF */



/* Finally, here's what we need!
 * Returns 0 if something goes wrong and it can't be found.
 * (UID 0 really shouldn't be logging in from a network anyway.)
 */
uid_t getUID (unsigned int localhost_ip,
              int  userport, int daemonport) {

#ifndef USE_AUTHLOCAL
  return (uid_t) 0;
#else
  uid_t retval;

  int fd = open(TCPSTAT_FNAME, O_RDONLY);
  if (fd == -1) {
    return (uid_t) 0;
  }
  retval = getUID_impl (localhost_ip, userport, daemonport, fd);
  close (fd);

  if (retval > AUTHLOCAL_MIN_UID)
    return retval;
  else
    return 0;

#endif	/* ifdef (USE_AUTHLOCAL) */
}


/* demonstration purposes.  Note that it's commented out.

		static struct in_addr localhost_ip;
		localhost_ip.s_addr = htonl(0x7f000001);
	...
		struct 	sockaddr_in user_address;
		int user_fd = accept
		  ( server_fd, (sockaddr *)&user_address, &len );
	...
		uid_t authenticated = 0;
		if (user_address.sin_addr.s_addr == localhost_ip.s_addr ) {

		Auto-Authenticate user if logging in from localhost!

		  authenticated = getUID
		    ( localhost_ip.s_addr, ntohs(user_address.sin_port),
		      server_portnum );
		}


int main(int argc, char **argv) {
  int pnum = 2048;
  if (argc > 1)
    pnum = (int) strtol(argv[1], (char **)NULL, 16);

  printf ("lookin' for a connection to port %d (0X%04X) from %d (0X%04X)\n",
	  5000, 5000, pnum, pnum);
  printf ("returned %d\n", (int) getUID ( inet_addr ("127.0.0.1"), pnum, 5000));
}
*/

